/*
 * warper (formerly trigger:warp_basic)
 *
 * [Warper]
 * Area0=x,y,w,h
 * Dest0=x,y,w,h
 *
 * Max 100 warps for each arena, but keep it low to minimise
 *  server load. Experiment to find out the best amount. Warps
 *  must be consecutive, fx: Area0, Area1, etc
 *
 * If you liked this module you'll love trigger_advanced. Trigger
 *  messages, prizes, ship changes, lvz and sound. Also featuring
 *  team dependant warps
 *
 * 25/06/03 - Created by Smong
 *
 * 27/06/03 - Removed some bogus FIX ME's. Added some interface
 *  checks to MM_func() that I also forgot to add.
 *
 * 28/06/03 - Memory is now allocated and freed as arenas
 *  attach/detach. ToDo: Add reload function when .conf changes.
 *
 * 23/01/04 - Changed rand to prng->Rand()
 *  Removed log messages, no longer has lm.
 *
 * 13/06/04 - Renamed from warp_basic2.c to warper.c, renaming module to warper.
 *  Lots of cleanups, now uses ppk callback. No longer an attach module, reads
 *  .conf to see if there is a valid [Warper] section.
 *
 */

#include <stdio.h>

#include "asss.h"
#include "region.h"

struct WarpData
{
	rect_t enter;
	rect_t exit;
};

typedef struct
{
	int count;
	struct WarpData *data;
} adata;

local void ArenaAction(Arena *arena, int action);
local void Pppk(Player *p, byte *p2, int len);

local void conf_extract_entries(ConfigHandle ch, adata *ad);
local int  conf_count_entries(ConfigHandle ch);

local void warp_player(Player *p, rect_t *r);
local int  check_rect(rect_t *r, struct C2SPosition *pos);

local int adkey;

local Imodman *mm;
local Iarenaman *aman;
local Iplayerdata *pd;
local Igame *game;
local Inet *net;
local Iconfig *cfg;
local Iprng *prng;

local void ArenaAction(Arena *arena, int action)
{
	adata *ad = P_ARENA_DATA(arena, adkey);

	if (action == AA_DESTROY || action == AA_CONFCHANGED)
	{
		if (ad->count > 0)
		{
			afree(ad->data);
			ad->count = 0;
		}
	}

	if (action == AA_CREATE || action == AA_CONFCHANGED)
	{
		ad->count = conf_count_entries(arena->cfg);
		if (ad->count > 0)
		{
			ad->data = amalloc( sizeof(struct WarpData) * ad->count );
			conf_extract_entries(arena->cfg, ad);
		}
	}
}

/* conf reader */

local void conf_extract_entries(ConfigHandle ch, adata *ad)
{
	int i;
	const char *value;
	char key[8];

	for(i=0; i<ad->count; i++)
	{
		snprintf(key, 8, "Area%d", i);
		value = cfg->GetStr(ch, "Warper", key);
		if (!value) return;

		sscanf(value, "%d,%d,%d,%d",
			&ad->data[i].enter.x, &ad->data[i].enter.y,
			&ad->data[i].enter.w, &ad->data[i].enter.h);

		snprintf(key, 8, "Dest%d", i);
		value = cfg->GetStr(ch, "Warper", key);
		if (!value) return;

		sscanf(value, "%d,%d,%d,%d",
			&ad->data[i].exit.x, &ad->data[i].exit.y,
			&ad->data[i].exit.w, &ad->data[i].exit.h);
	}
}

local int conf_count_entries(ConfigHandle ch)
{
	int i;
	const char *value;
	char key[8];

	for(i=0; i<100; i++)
	{
		snprintf(key, 8, "Area%d", i);
		value = cfg->GetStr(ch, "Warper", key);
		if (!value) break;

		snprintf(key, 8, "Dest%d", i);
		value = cfg->GetStr(ch, "Warper", key);
		if (!value) break;
	}

	return i;
}

/* warp code */

local void warp_player(Player *p, rect_t *r)
{
	Target tgt;
	int tx, ty;

	tx = r->x;
	ty = r->y;
	if (r->w > 0) tx += prng->Rand() % r->w;
	if (r->h > 0) ty += prng->Rand() % r->h;

	tgt.type = T_PLAYER;
	tgt.u.p  = p;
	game->WarpTo(&tgt, tx, ty);
}

local int check_rect(rect_t *r, struct C2SPosition *pos)
{
	int x = pos->x >> 4;
	int y = pos->y >> 4;

	if (x < r->x) return 0;
	if (y < r->y) return 0;
	if (x > r->x + r->w) return 0;
	if (y > r->y + r->h) return 0;

	return 1;
}

local void Pppk(Player *p, byte *p2, int len)
{
	struct C2SPosition *pos = (struct C2SPosition *)p2;
	Arena *arena = p->arena;
	adata *ad = P_ARENA_DATA(arena, adkey);
	int i, k;

	/* sanity check */
	if (len < 22) return;
	if (!arena) return;
	if (ad->count < 1) return;
	if (p->p_ship == SPEC) return;

	k = ad->count;
	for(i=0; i<k; i++)
		if (check_rect(&ad->data[i].enter, pos))
			break;
	if (i >= k) return;

	warp_player(p, &ad->data[i].exit);
}

EXPORT const char info_warper[] = "v1";

EXPORT int MM_warper(int action, Imodman *mm_, Arena *arena)
{
	if (action == MM_LOAD)
	{
		mm = mm_;
		aman = mm->GetInterface(I_ARENAMAN, ALLARENAS);
		pd = mm->GetInterface(I_PLAYERDATA, ALLARENAS);
		game = mm->GetInterface(I_GAME, ALLARENAS);
		net = mm->GetInterface(I_NET, ALLARENAS);
		cfg = mm->GetInterface(I_CONFIG, ALLARENAS);
		prng = mm->GetInterface(I_PRNG, ALLARENAS);
		if (!aman || !pd || !game || !net || !cfg || !prng)
			return MM_FAIL;

		adkey = aman->AllocateArenaData( sizeof(adata) );
		if (adkey == -1) return MM_FAIL;

		mm->RegCallback(CB_ARENAACTION, ArenaAction, ALLARENAS);
		net->AddPacket(C2S_POSITION, Pppk);
		return MM_OK;
	}
	else if (action == MM_UNLOAD)
	{
		mm->UnregCallback(CB_ARENAACTION, ArenaAction, ALLARENAS);
		net->RemovePacket(C2S_POSITION, Pppk);
		aman->FreeArenaData(adkey);

		mm->ReleaseInterface(aman);
		mm->ReleaseInterface(pd);
		mm->ReleaseInterface(game);
		mm->ReleaseInterface(net);
		mm->ReleaseInterface(cfg);
		mm->ReleaseInterface(prng);
		return MM_OK;
	}

	return MM_FAIL;
}

